@@ -1,25 +1,7 @@
 /*
- * Copyright (C) 2021 Valve Corporation
- * Copyright (C) 2014 Rob Clark <robclark@freedesktop.org>
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
- * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
- * SOFTWARE.
+ * Copyright © 2021 Valve Corporation
+ * Copyright © 2014 Rob Clark <robclark@freedesktop.org>
+ * SPDX-License-Identifier: MIT
  */
 
 #include "ir3_ra.h"
@@ -366,7 +348,7 @@ struct ra_ctx {
    struct ir3_block *block;
 
    const struct ir3_compiler *compiler;
-   gl_shader_stage stage;
+   mesa_shader_stage stage;
 
    /* Pending moves of top-level intervals that will be emitted once we're
     * finished:
@@ -404,6 +386,13 @@ rb_node_to_interval_const(const struct rb_node *node)
    return rb_node_data(struct ra_interval, node, physreg_node);
 }
 
+static struct ra_interval *
+ra_interval_get(struct ra_ctx *ctx, struct ir3_register *dst)
+{
+   assert(dst->name != 0 && dst->name < ctx->live->definitions_count);
+   return &ctx->intervals[dst->name];
+}
+
 static struct ra_interval *
 ra_interval_next(struct ra_interval *interval)
 {
@@ -773,7 +762,7 @@ check_dst_overlap(struct ra_ctx *ctx, struct ra_file *file,
       if (ra_get_file(ctx, other_dst) != file)
          continue;
 
-      struct ra_interval *other_interval = &ctx->intervals[other_dst->name];
+      struct ra_interval *other_interval = ra_interval_get(ctx, other_dst);
       assert(!other_interval->interval.parent);
       physreg_t other_start = other_interval->physreg_start;
       physreg_t other_end = other_interval->physreg_end;
@@ -1093,7 +1082,7 @@ compress_regs_left(struct ra_ctx *ctx, struct ra_file *file,
          if (dst_inserted[n])
             continue;
 
-         struct ra_interval *other_interval = &ctx->intervals[other_dst->name];
+         struct ra_interval *other_interval = ra_interval_get(ctx, other_dst);
          /* if the destination partially overlaps this interval, we need to
           * extend candidate_start to the end.
           */
@@ -1111,7 +1100,7 @@ compress_regs_left(struct ra_ctx *ctx, struct ra_file *file,
           */
          if (other_dst->tied) {
             struct ra_interval *tied_interval =
-               &ctx->intervals[other_dst->tied->def->name];
+               ra_interval_get(ctx, other_dst->tied->def);
             if (tied_interval->is_killed)
                continue;
          }
@@ -1248,7 +1237,7 @@ compress_regs_left(struct ra_ctx *ctx, struct ra_file *file,
       }
 
       if (!(cur_reg->flags & IR3_REG_HALF))
-         physreg = ALIGN(physreg, 2);
+         physreg = align(physreg, 2);
 
       d("pushing reg %u physreg %u\n", cur_reg->name, physreg);
 
@@ -1257,7 +1246,7 @@ compress_regs_left(struct ra_ctx *ctx, struct ra_file *file,
           reg_file_size(file, cur_reg)) {
          d("ran out of room for interval %u!\n",
            cur_reg->name);
-         unreachable("reg pressure calculation was wrong!");
+         UNREACHABLE("reg pressure calculation was wrong!");
          return 0;
       }
 
@@ -1265,7 +1254,7 @@ compress_regs_left(struct ra_ctx *ctx, struct ra_file *file,
          if (cur_reg == reg) {
             ret_reg = physreg;
          } else {
-            struct ra_interval *interval = &ctx->intervals[cur_reg->name];
+            struct ra_interval *interval = ra_interval_get(ctx, cur_reg);
             interval->physreg_start = physreg;
             interval->physreg_end = physreg + interval_size;
          }
@@ -1295,11 +1284,11 @@ compress_regs_left(struct ra_ctx *ctx, struct ra_file *file,
       if (!tied)
          continue;
 
-      struct ra_interval *tied_interval = &ctx->intervals[tied->def->name];
+      struct ra_interval *tied_interval = ra_interval_get(ctx, tied->def);
       if (!tied_interval->is_killed)
          continue;
 
-      struct ra_interval *dst_interval = &ctx->intervals[dst->name];
+      struct ra_interval *dst_interval = ra_interval_get(ctx, dst);
       unsigned dst_size = reg_size(dst);
       dst_interval->physreg_start = ra_interval_get_physreg(tied_interval);
       dst_interval->physreg_end = dst_interval->physreg_start + dst_size;
@@ -1341,7 +1330,9 @@ find_best_gap(struct ra_ctx *ctx, struct ra_file *file,
    BITSET_WORD *available =
       is_early_clobber(dst) ? file->available_to_evict : file->available;
 
-   unsigned start = ALIGN(file->start, alignment) % (file_size - size + alignment);
+   unsigned start = align(file->start, alignment);
+   if (start + size > file_size)
+      start = 0;
    unsigned candidate = start;
    do {
       bool is_available = true;
@@ -1370,6 +1361,110 @@ find_best_gap(struct ra_ctx *ctx, struct ra_file *file,
    return (physreg_t)~0;
 }
 
+static physreg_t
+try_allocate_src(struct ra_ctx *ctx, struct ra_file *file,
+                 struct ir3_register *reg)
+{
+   unsigned file_size = reg_file_size(file, reg);
+   unsigned size = reg_size(reg);
+   for (unsigned i = 0; i < reg->instr->srcs_count; i++) {
+      struct ir3_register *src = reg->instr->srcs[i];
+      if (!ra_reg_is_src(src))
+         continue;
+      if (ra_get_file(ctx, src) == file && reg_size(src) >= size) {
+         struct ra_interval *src_interval = ra_interval_get(ctx, src->def);
+         physreg_t src_physreg = ra_interval_get_physreg(src_interval);
+         if (src_physreg % reg_elem_size(reg) == 0 &&
+             src_physreg + size <= file_size &&
+             get_reg_specified(ctx, file, reg, src_physreg, false))
+            return src_physreg;
+      }
+   }
+
+   return ~0;
+}
+
+static physreg_t
+try_allocate_src_subreg(struct ra_ctx *ctx, struct ra_file *file,
+                        struct ir3_register *reg,
+                        enum ir3_subreg_move subreg_move)
+{
+   assert(subreg_move != IR3_SUBREG_MOVE_NONE);
+
+   /* Subreg moves always write a half register. */
+   assert(reg_elem_size(reg) == 1);
+
+   struct ir3_register *src = reg->instr->srcs[0];
+   if (!ra_reg_is_src(src) || ra_get_file(ctx, src) != file)
+      return ~0;
+
+   unsigned offset = subreg_move == IR3_SUBREG_MOVE_LOWER ? 0 : 1;
+   struct ra_interval *src_interval = ra_interval_get(ctx, src->def);
+   physreg_t src_physreg = ra_interval_get_physreg(src_interval) + offset;
+   unsigned file_size = reg_file_size(file, reg);
+   unsigned size = reg_size(reg);
+
+   if (src_physreg + size <= file_size &&
+       get_reg_specified(ctx, file, reg, src_physreg, false)) {
+      return src_physreg;
+   }
+
+   return ~0;
+}
+
+static bool
+rpt_has_unique_merge_set(struct ir3_instruction *instr)
+{
+   assert(ir3_instr_is_rpt(instr));
+
+   if (!instr->dsts[0]->merge_set)
+      return false;
+
+   struct ir3_instruction *first = ir3_instr_first_rpt(instr);
+   struct ir3_register *def = first->dsts[0];
+
+   if (def->merge_set != instr->dsts[0]->merge_set ||
+       def->merge_set->regs_count != ir3_instr_rpt_length(first)) {
+      return false;
+   }
+
+   unsigned i = 0;
+
+   foreach_instr_rpt (rpt, first) {
+      if (rpt->dsts[0] != def->merge_set->regs[i++])
+         return false;
+   }
+
+   return true;
+}
+
+/* Handles this case when a reg's merge set has a preferred reg but is currently
+ * unavailable. In this case, it's often preferable to reset its preferred reg
+ * and assign a new one, as this potentially reduces the number of movs needed
+ * for the as of yet unallocated regs.
+ */
+void
+ir3_ra_handle_unavailable_merge_set(struct ir3_register *reg)
+{
+   unsigned num_unallocated = 0;
+
+   for (unsigned i = 0; i < reg->merge_set->regs_count; i++) {
+      if (reg->merge_set->regs[i]->num == INVALID_REG) {
+         num_unallocated++;
+
+         /* Only reset the preferred reg if there are at least two still
+          * unallocated regs. It doesn't make sense to reassign the merge set
+          * for a single reg, and increasing the bound more doesn't seem to
+          * improve shader stats.
+          */
+         if (num_unallocated >= 2) {
+            reg->merge_set->preferred_reg = (physreg_t)~0;
+            return;
+         }
+      }
+   }
+}
+
 /* This is the main entrypoint for picking a register. Pick a free register
  * for "reg", shuffling around sources if necessary. In the normal case where
  * "is_source" is false, this register can overlap with killed sources
@@ -1382,6 +1477,16 @@ find_best_gap(struct ra_ctx *ctx, struct ra_file *file,
 static physreg_t
 get_reg(struct ra_ctx *ctx, struct ra_file *file, struct ir3_register *reg)
 {
+   /* For subreg moves (see ir3_is_subreg_move), try to allocate half of their
+    * full src for their dst. If this succeeds, the instruction can be removed.
+    */
+   enum ir3_subreg_move subreg_move = ir3_is_subreg_move(reg->instr);
+   if (subreg_move != IR3_SUBREG_MOVE_NONE) {
+      physreg_t src_reg = try_allocate_src_subreg(ctx, file, reg, subreg_move);
+      if (src_reg != (physreg_t)~0)
+         return src_reg;
+   }
+
    unsigned file_size = reg_file_size(file, reg);
    if (reg->merge_set && reg->merge_set->preferred_reg != (physreg_t)~0) {
       physreg_t preferred_reg =
@@ -1390,6 +1495,20 @@ get_reg(struct ra_ctx *ctx, struct ra_file *file, struct ir3_register *reg)
           preferred_reg % reg_elem_size(reg) == 0 &&
           get_reg_specified(ctx, file, reg, preferred_reg, false))
          return preferred_reg;
+
+      ir3_ra_handle_unavailable_merge_set(reg);
+   }
+
+   /* For repeated instructions whose merge set is unique (i.e., only used for
+    * these repeated instructions), try to first allocate one of their sources
+    * (for the same reason as for ALU/SFU instructions explained below). This
+    * also prevents us from allocating a new register range for this merge set
+    * when the one from a source could be reused.
+    */
+   if (ir3_instr_is_rpt(reg->instr) && rpt_has_unique_merge_set(reg->instr)) {
+      physreg_t src_reg = try_allocate_src(ctx, file, reg);
+      if (src_reg != (physreg_t)~0)
+         return src_reg;
    }
 
    /* If this register is a subset of a merge set which we have not picked a
@@ -1414,19 +1533,9 @@ get_reg(struct ra_ctx *ctx, struct ra_file *file, struct ir3_register *reg)
     * SFU instructions:
     */
    if (is_sfu(reg->instr) || is_alu(reg->instr)) {
-      for (unsigned i = 0; i < reg->instr->srcs_count; i++) {
-         struct ir3_register *src = reg->instr->srcs[i];
-         if (!ra_reg_is_src(src))
-            continue;
-         if (ra_get_file(ctx, src) == file && reg_size(src) >= size) {
-            struct ra_interval *src_interval = &ctx->intervals[src->def->name];
-            physreg_t src_physreg = ra_interval_get_physreg(src_interval);
-            if (src_physreg % reg_elem_size(reg) == 0 &&
-                src_physreg + size <= file_size &&
-                get_reg_specified(ctx, file, reg, src_physreg, false))
-               return src_physreg;
-         }
-      }
+      physreg_t src_reg = try_allocate_src(ctx, file, reg);
+      if (src_reg != (physreg_t)~0)
+         return src_reg;
    }
 
    physreg_t best_reg =
@@ -1479,7 +1588,7 @@ assign_reg(struct ir3_instruction *instr, struct ir3_register *reg,
 static void
 mark_src_killed(struct ra_ctx *ctx, struct ir3_register *src)
 {
-   struct ra_interval *interval = &ctx->intervals[src->def->name];
+   struct ra_interval *interval = ra_interval_get(ctx, src->def);
 
    if (!(src->flags & IR3_REG_FIRST_KILL) || interval->is_killed ||
        interval->interval.parent ||
@@ -1493,7 +1602,7 @@ static void
 insert_dst(struct ra_ctx *ctx, struct ir3_register *dst)
 {
    struct ra_file *file = ra_get_file(ctx, dst);
-   struct ra_interval *interval = &ctx->intervals[dst->name];
+   struct ra_interval *interval = ra_interval_get(ctx, dst);
 
    d("insert dst %u physreg %u", dst->name, ra_interval_get_physreg(interval));
 
@@ -1508,7 +1617,7 @@ allocate_dst_fixed(struct ra_ctx *ctx, struct ir3_register *dst,
                    physreg_t physreg)
 {
    struct ra_file *file = ra_get_file(ctx, dst);
-   struct ra_interval *interval = &ctx->intervals[dst->name];
+   struct ra_interval *interval = ra_interval_get(ctx, dst);
    ra_update_affinity(file->size, dst, physreg);
 
    ra_interval_init(interval, dst);
@@ -1533,8 +1642,8 @@ insert_tied_dst_copy(struct ra_ctx *ctx, struct ir3_register *dst)
    if (!tied)
       return;
 
-   struct ra_interval *tied_interval = &ctx->intervals[tied->def->name];
-   struct ra_interval *dst_interval = &ctx->intervals[dst->name];
+   struct ra_interval *tied_interval = ra_interval_get(ctx, tied->def);
+   struct ra_interval *dst_interval = ra_interval_get(ctx, dst);
 
    if (tied_interval->is_killed)
       return;
@@ -1555,7 +1664,7 @@ allocate_dst(struct ra_ctx *ctx, struct ir3_register *dst)
 
    struct ir3_register *tied = dst->tied;
    if (tied) {
-      struct ra_interval *tied_interval = &ctx->intervals[tied->def->name];
+      struct ra_interval *tied_interval = ra_interval_get(ctx, tied->def);
       if (tied_interval->is_killed) {
          /* The easy case: the source is killed, so we can just reuse it
           * for the destination.
@@ -1575,13 +1684,13 @@ static void
 assign_src(struct ra_ctx *ctx, struct ir3_instruction *instr,
            struct ir3_register *src)
 {
-   struct ra_interval *interval = &ctx->intervals[src->def->name];
+   struct ra_interval *interval = ra_interval_get(ctx, src->def);
    struct ra_file *file = ra_get_file(ctx, src);
 
    struct ir3_register *tied = src->tied;
    physreg_t physreg;
    if (tied) {
-      struct ra_interval *tied_interval = &ctx->intervals[tied->name];
+      struct ra_interval *tied_interval = ra_interval_get(ctx, tied);
       physreg = ra_interval_get_physreg(tied_interval);
    } else {
       physreg = ra_interval_get_physreg(interval);
@@ -1602,9 +1711,9 @@ insert_parallel_copy_instr(struct ra_ctx *ctx, struct ir3_instruction *instr)
    if (ctx->parallel_copies_count == 0)
       return;
 
-   struct ir3_instruction *pcopy =
-      ir3_instr_create(instr->block, OPC_META_PARALLEL_COPY,
-                       ctx->parallel_copies_count, ctx->parallel_copies_count);
+   struct ir3_instruction *pcopy = ir3_instr_create_at(
+      ir3_before_instr(instr), OPC_META_PARALLEL_COPY,
+      ctx->parallel_copies_count, ctx->parallel_copies_count);
 
    for (unsigned i = 0; i < ctx->parallel_copies_count; i++) {
       struct ra_parallel_copy *entry = &ctx->parallel_copies[i];
@@ -1628,8 +1737,6 @@ insert_parallel_copy_instr(struct ra_ctx *ctx, struct ir3_instruction *instr)
       assign_reg(pcopy, reg, ra_physreg_to_num(entry->src, reg->flags));
    }
 
-   list_del(&pcopy->node);
-   list_addtail(&pcopy->node, &instr->node);
    ctx->parallel_copies_count = 0;
 }
 
@@ -1681,7 +1788,7 @@ handle_split(struct ra_ctx *ctx, struct ir3_instruction *instr)
       return;
    }
 
-   struct ra_interval *src_interval = &ctx->intervals[src->def->name];
+   struct ra_interval *src_interval = ra_interval_get(ctx, src->def);
 
    physreg_t physreg = ra_interval_get_physreg(src_interval);
    assign_src(ctx, instr, src);
@@ -1727,7 +1834,7 @@ handle_collect(struct ra_ctx *ctx, struct ir3_instruction *instr)
          mark_src_killed(ctx, src);
       }
 
-      struct ra_interval *interval = &ctx->intervals[src->def->name];
+      struct ra_interval *interval = ra_interval_get(ctx, src->def);
 
       /* We only need special handling if the source's interval overlaps with
        * the destination's interval.
@@ -1761,13 +1868,13 @@ handle_collect(struct ra_ctx *ctx, struct ir3_instruction *instr)
 
    /* Remove the temporary is_killed we added */
    ra_foreach_src (src, instr) {
-      struct ra_interval *interval = &ctx->intervals[src->def->name];
+      struct ra_interval *interval = ra_interval_get(ctx, src->def);
       while (interval->interval.parent != NULL) {
          interval = ir3_reg_interval_to_ra_interval(interval->interval.parent);
       }
 
       /* Filter out cases where it actually should be killed */
-      if (interval != &ctx->intervals[src->def->name] ||
+      if (interval != ra_interval_get(ctx, src->def) ||
           !(src->flags & IR3_REG_KILL)) {
          ra_file_unmark_killed(ra_get_file(ctx, src), interval);
       }
@@ -1823,7 +1930,7 @@ handle_precolored_input(struct ra_ctx *ctx, struct ir3_instruction *instr)
       return;
 
    struct ra_file *file = ra_get_file(ctx, instr->dsts[0]);
-   struct ra_interval *interval = &ctx->intervals[instr->dsts[0]->name];
+   struct ra_interval *interval = ra_interval_get(ctx, instr->dsts[0]);
    physreg_t physreg = ra_reg_get_physreg(instr->dsts[0]);
    allocate_dst_fixed(ctx, instr->dsts[0], physreg);
 
@@ -1843,7 +1950,7 @@ handle_input(struct ra_ctx *ctx, struct ir3_instruction *instr)
    allocate_dst(ctx, instr->dsts[0]);
 
    struct ra_file *file = ra_get_file(ctx, instr->dsts[0]);
-   struct ra_interval *interval = &ctx->intervals[instr->dsts[0]->name];
+   struct ra_interval *interval = ra_interval_get(ctx, instr->dsts[0]);
    ra_file_insert(file, interval);
 }
 
@@ -1853,7 +1960,7 @@ assign_input(struct ra_ctx *ctx, struct ir3_instruction *instr)
    if (!(instr->dsts[0]->flags & IR3_REG_SSA))
       return;
 
-   struct ra_interval *interval = &ctx->intervals[instr->dsts[0]->name];
+   struct ra_interval *interval = ra_interval_get(ctx, instr->dsts[0]);
    struct ra_file *file = ra_get_file(ctx, instr->dsts[0]);
 
    if (instr->dsts[0]->num == INVALID_REG) {
@@ -1889,7 +1996,7 @@ static void
 handle_precolored_source(struct ra_ctx *ctx, struct ir3_register *src)
 {
    struct ra_file *file = ra_get_file(ctx, src);
-   struct ra_interval *interval = &ctx->intervals[src->def->name];
+   struct ra_interval *interval = ra_interval_get(ctx, src->def);
    physreg_t physreg = ra_reg_get_physreg(src);
 
    if (ra_interval_get_num(interval) == src->num)
@@ -1903,7 +2010,7 @@ handle_precolored_source(struct ra_ctx *ctx, struct ir3_register *src)
       unsigned eviction_count;
       if (!try_evict_regs(ctx, file, src, physreg, &eviction_count, true,
                           false)) {
-         unreachable("failed to evict for precolored source!");
+         UNREACHABLE("failed to evict for precolored source!");
          return;
       }
    }
@@ -1926,7 +2033,7 @@ handle_chmask(struct ra_ctx *ctx, struct ir3_instruction *instr)
 
    ra_foreach_src (src, instr) {
       struct ra_file *file = ra_get_file(ctx, src);
-      struct ra_interval *interval = &ctx->intervals[src->def->name];
+      struct ra_interval *interval = ra_interval_get(ctx, src->def);
       if (src->flags & IR3_REG_FIRST_KILL)
          ra_file_remove(file, interval);
    }
@@ -1966,7 +2073,7 @@ handle_live_in(struct ra_ctx *ctx, struct ir3_register *def)
 
    assert(physreg != (physreg_t)~0);
 
-   struct ra_interval *interval = &ctx->intervals[def->name];
+   struct ra_interval *interval = ra_interval_get(ctx, def);
    struct ra_file *file = ra_get_file(ctx, def);
    ra_interval_init(interval, def);
    interval->physreg_start = physreg;
@@ -1985,7 +2092,7 @@ handle_live_out(struct ra_ctx *ctx, struct ir3_register *def)
       return;
 
    struct ra_block_state *state = &ctx->blocks[ctx->block->index];
-   struct ra_interval *interval = &ctx->intervals[def->name];
+   struct ra_interval *interval = ra_interval_get(ctx, def);
    physreg_t physreg = ra_interval_get_physreg(interval);
    if (physreg != ra_reg_get_physreg(def)) {
       if (!state->renames)
@@ -1997,11 +2104,8 @@ handle_live_out(struct ra_ctx *ctx, struct ir3_register *def)
 static void
 handle_phi(struct ra_ctx *ctx, struct ir3_register *def)
 {
-   if (!(def->flags & IR3_REG_SSA))
-      return;
-
    struct ra_file *file = ra_get_file(ctx, def);
-   struct ra_interval *interval = &ctx->intervals[def->name];
+   struct ra_interval *interval = ra_interval_get(ctx, def);
 
    /* phis are always scalar, so they should already be the smallest possible
     * size. However they may be coalesced with other live-in values/phi
@@ -2030,7 +2134,7 @@ assign_phi(struct ra_ctx *ctx, struct ir3_instruction *phi)
       return;
 
    struct ra_file *file = ra_get_file(ctx, phi->dsts[0]);
-   struct ra_interval *interval = &ctx->intervals[phi->dsts[0]->name];
+   struct ra_interval *interval = ra_interval_get(ctx, phi->dsts[0]);
    assert(!interval->interval.parent);
    unsigned num = ra_interval_get_num(interval);
    assign_reg(phi, phi->dsts[0], num);
@@ -2080,8 +2184,9 @@ insert_liveout_copy(struct ir3_block *block, physreg_t dst, physreg_t src,
       old_pcopy = last;
 
    unsigned old_pcopy_srcs = old_pcopy ? old_pcopy->srcs_count : 0;
-   struct ir3_instruction *pcopy = ir3_instr_create(
-      block, OPC_META_PARALLEL_COPY, old_pcopy_srcs + 1, old_pcopy_srcs + 1);
+   struct ir3_instruction *pcopy =
+      ir3_instr_create_at(ir3_before_terminator(block), OPC_META_PARALLEL_COPY,
+                          old_pcopy_srcs + 1, old_pcopy_srcs + 1);
 
    for (unsigned i = 0; i < old_pcopy_srcs; i++) {
       old_pcopy->dsts[i]->instr = pcopy;
@@ -2224,6 +2329,17 @@ insert_live_out_moves(struct ra_ctx *ctx)
    insert_file_live_out_moves(ctx, &ctx->shared);
 }
 
+static bool
+has_merge_set_preferred_reg(struct ir3_register *reg)
+{
+   assert(reg->merge_set);
+   assert(reg->num != INVALID_REG);
+
+   return reg->merge_set->preferred_reg != (physreg_t)~0 &&
+          ra_reg_get_physreg(reg) ==
+             reg->merge_set->preferred_reg + reg->merge_set_offset;
+}
+
 static void
 handle_block(struct ra_ctx *ctx, struct ir3_block *block)
 {
@@ -2234,6 +2350,50 @@ handle_block(struct ra_ctx *ctx, struct ir3_block *block)
    ra_file_init(&ctx->half);
    ra_file_init(&ctx->shared);
 
+   if (block == ir3_after_preamble(block->shader) &&
+       block != ir3_start_block(block->shader)) {
+      /* Reset the file start in the first block after the preamble to make the
+       * main shader independent of the preamble. Without this, the allocated
+       * registers in the main shader will depend on how many registers were
+       * used in the preamble. This in turn may cause more or less copies being
+       * generated or postsched behaving differently due to a difference in
+       * false dependencies. This is undesirable when analyzing compiler changes
+       * that should only affect the preamble as they may also change main
+       * shader stats, generating noise in the shader-db output.
+       */
+      ctx->full.start = 0;
+      ctx->half.start = 0;
+      ctx->shared.start = 0;
+
+      /* However, make sure the file start accounts for defs that are
+       * live-through the preamble (inputs and tex prefetches). If not, this
+       * could introduce unwanted false dependencies.
+       */
+      foreach_instr (input, &ir3_start_block(block->shader)->instr_list) {
+         if (input->opc != OPC_META_INPUT &&
+             input->opc != OPC_META_TEX_PREFETCH) {
+            break;
+         }
+
+         struct ir3_register *dst = input->dsts[0];
+         assert(dst->num != INVALID_REG);
+
+         physreg_t dst_end;
+
+         if (dst->merge_set && has_merge_set_preferred_reg(dst)) {
+            /* Take the whole merge set into account to prevent its range being
+             * allocated for defs not part of the merge set.
+             */
+            dst_end = dst->merge_set->preferred_reg + dst->merge_set->size;
+         } else {
+            dst_end = ra_reg_get_physreg(dst) + reg_size(dst);
+         }
+
+         struct ra_file *file = ra_get_file(ctx, dst);
+         file->start = MAX2(file->start, dst_end);
+      }
+   }
+
    /* Handle live-ins, phis, and input meta-instructions. These all appear
     * live at the beginning of the block, and interfere with each other
     * therefore need to be allocated "in parallel". This means that we
@@ -2261,14 +2421,51 @@ handle_block(struct ra_ctx *ctx, struct ir3_block *block)
       handle_live_in(ctx, reg);
    }
 
+   /* Handle phis in two groups: first those which already have a preferred reg
+    * set and then those without. The second group should be rare but by
+    * handling them last, they don't accidentally occupy a preferred reg of
+    * another phi, preventing excessive copying in some cases.
+    */
+   bool skipped_phi = false;
+
    foreach_instr (instr, &block->instr_list) {
-      if (instr->opc == OPC_META_PHI)
-         handle_phi(ctx, instr->dsts[0]);
-      else if (instr->opc == OPC_META_INPUT ||
-               instr->opc == OPC_META_TEX_PREFETCH)
+      if (instr->opc == OPC_META_PHI) {
+         struct ir3_register *dst = instr->dsts[0];
+
+         /* Some phis may have been handled by shared RA already. */
+         if (!(dst->flags & IR3_REG_SSA)) {
+            continue;
+         }
+
+         if (dst->merge_set && dst->merge_set->preferred_reg != (physreg_t)~0) {
+            handle_phi(ctx, dst);
+         } else {
+            skipped_phi = true;
+         }
+      } else if (instr->opc == OPC_META_INPUT ||
+                 instr->opc == OPC_META_TEX_PREFETCH) {
          handle_input(ctx, instr);
-      else
+      } else {
          break;
+      }
+   }
+
+   if (skipped_phi) {
+      foreach_instr (instr, &block->instr_list) {
+         if (instr->opc == OPC_META_PHI) {
+            struct ir3_register *dst = instr->dsts[0];
+
+            if (!(dst->flags & IR3_REG_SSA)) {
+               continue;
+            }
+
+            if (!ra_interval_get(ctx, dst)->interval.inserted) {
+               handle_phi(ctx, dst);
+            }
+         } else {
+            break;
+         }
+      }
    }
 
    /* After this point, every live-in/phi/input has an interval assigned to
@@ -2506,10 +2703,16 @@ calc_limit_pressure_for_cs_with_barrier(struct ir3_shader_variant *v,
 {
    const struct ir3_compiler *compiler = v->compiler;
 
+   bool double_threadsize = ir3_should_double_threadsize(v, 0);
    unsigned threads_per_wg;
+
    if (v->local_size_variable) {
-      /* We have to expect the worst case. */
-      threads_per_wg = compiler->max_variable_workgroup_size;
+      if (v->type == MESA_SHADER_KERNEL) {
+         threads_per_wg = compiler->threadsize_base * (double_threadsize ? 2 : 1);
+      } else {
+         /* We have to expect the worst case. */
+         threads_per_wg = compiler->max_variable_workgroup_size;
+      }
    } else {
       threads_per_wg = v->local_size[0] * v->local_size[1] * v->local_size[2];
    }
@@ -2522,7 +2725,6 @@ calc_limit_pressure_for_cs_with_barrier(struct ir3_shader_variant *v,
     * parts each could get.
     */
 
-   bool double_threadsize = ir3_should_double_threadsize(v, 0);
    unsigned waves_per_wg = DIV_ROUND_UP(
       threads_per_wg, compiler->threadsize_base * (double_threadsize ? 2 : 1) *
                          compiler->wave_granularity);
@@ -2544,6 +2746,33 @@ calc_limit_pressure_for_cs_with_barrier(struct ir3_shader_variant *v,
    }
 }
 
+struct ir3_pressure
+ir3_ra_get_reg_file_limits(struct ir3_shader_variant *v)
+{
+   struct ir3_pressure limit_pressure = {
+      .full = RA_FULL_SIZE,
+      .half = RA_HALF_SIZE,
+      .shared = RA_SHARED_SIZE,
+      .shared_half = RA_SHARED_HALF_SIZE,
+   };
+
+   if (mesa_shader_stage_is_compute(v->type) &&
+       v->shader->nir->info.uses_control_barrier) {
+      calc_limit_pressure_for_cs_with_barrier(v, &limit_pressure);
+   }
+
+   /* If the user forces a doubled threadsize, we may have to lower the limit
+    * because on some gens the register file is not big enough to hold a
+    * double-size wave with all 48 registers in use.
+    */
+   if (v->shader_options.real_wavesize == IR3_DOUBLE_ONLY) {
+      limit_pressure.full =
+         MAX2(limit_pressure.full, v->compiler->reg_size_vec4 / 2 * 16);
+   }
+
+   return limit_pressure;
+}
+
 int
 ir3_ra(struct ir3_shader_variant *v)
 {
@@ -2586,24 +2815,7 @@ ir3_ra(struct ir3_shader_variant *v)
    d("\thalf: %u", max_pressure.half);
    d("\tshared: %u", max_pressure.shared);
 
-   struct ir3_pressure limit_pressure;
-   limit_pressure.full = RA_FULL_SIZE;
-   limit_pressure.half = RA_HALF_SIZE;
-   limit_pressure.shared = RA_SHARED_SIZE;
-   limit_pressure.shared_half = RA_SHARED_HALF_SIZE;
-
-   if (gl_shader_stage_is_compute(v->type) && v->has_barrier) {
-      calc_limit_pressure_for_cs_with_barrier(v, &limit_pressure);
-   }
-
-   /* If the user forces a doubled threadsize, we may have to lower the limit
-    * because on some gens the register file is not big enough to hold a
-    * double-size wave with all 48 registers in use.
-    */
-   if (v->shader_options.real_wavesize == IR3_DOUBLE_ONLY) {
-      limit_pressure.full =
-         MAX2(limit_pressure.full, ctx->compiler->reg_size_vec4 / 2 * 16);
-   }
+   struct ir3_pressure limit_pressure = ir3_ra_get_reg_file_limits(v);
 
    /* If requested, lower the limit so that spilling happens more often. */
    if (ir3_shader_debug & IR3_DBG_SPILLALL)
